#include <stdexcept>
#include <cstdio>
#include <ctime>
#include "gldbg.h"
#include "profiler.h"

using namespace std;

Profiler::Profiler(GlDbg &dbg):
	decoder(gldecoder_new(this, 0)),
	enabled(false)
{
	dbg.get_command_interpreter().register_command("profile", this, &Profiler::cmd_profile)
		.set_help("Profiles GL usage and performance",
			"profile {on|off}\n"
			"  Enables or disables profiling\n");

	decoder->glDrawArrays = glDrawArrays;
	decoder->glDrawElements = glDrawElements;
	decoder->glDrawRangeElements = glDrawRangeElements;
	decoder->glXSwapBuffers = glXSwapBuffers;
}

void Profiler::decode(const char *data, unsigned len)
{
	if(enabled)
		gldecoder_decode(decoder, data, len);
}

void Profiler::cmd_profile(const string &args)
{
	if(args.empty() || args=="on")
	{
		enabled = true;
		frames = 0;
		draw_calls = 0;
		vertices = 0;
		triangles = 0;
		start = get_time();
	}
	else if(args=="off")
		enabled = false;
	else
		throw runtime_error("Invalid argument");
}

void Profiler::glDrawArrays(void *user_data, GLenum mode, int, int count)
{
	glDrawElements(user_data, mode, count, 0, 0);
}

void Profiler::glDrawElements(void *user_data, GLenum mode, int count, GLenum, const void *)
{
	Profiler *self = reinterpret_cast<Profiler *>(user_data);

	++self->draw_calls;
	self->vertices += count;

	switch(mode)
	{
	case GL_TRIANGLES:
		self->triangles += count/3;
		break;
	case GL_TRIANGLE_STRIP:
	case GL_TRIANGLE_FAN:
		if(count>2)
			self->triangles += count-2;
		break;
	case GL_QUADS:
		self->triangles += count/4;
		break;
	case GL_QUAD_STRIP:
		if(count>3)
			self->triangles += count/2-1;
		break;
	}
}

void Profiler::glDrawRangeElements(void *user_data, GLenum mode, unsigned, unsigned, int count, GLenum type, const void *data)
{
	glDrawElements(user_data, mode, count, type, data);
}

void Profiler::glXSwapBuffers(void *user_data, Display *, GLXDrawable)
{
	Profiler *self = reinterpret_cast<Profiler *>(user_data);

	printf("%d draw calls, %d vertices, %d triangles\n", self->draw_calls, self->vertices, self->triangles);
	self->draw_calls = 0;
	self->vertices = 0;
	self->triangles = 0;

	++self->frames;
	Time t = get_time();
	float dt = (t-self->start)/1e6;
	if(dt>1)
	{
		printf("%d frames in %.2f seconds\n", self->frames, dt);
		self->start = t;
		self->frames = 0;
	}
}

Profiler::Time Profiler::get_time()
{
	struct timespec ts;
	clock_gettime(CLOCK_REALTIME, &ts);
	return ts.tv_sec*1000000+ts.tv_nsec/1000;
}
